// Copyright 2019 The TCMalloc Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "absl/types/span.h"
#include "tcmalloc/common.h"
#include "tcmalloc/internal/config.h"
#include "tcmalloc/size_class_info.h"
#include "tcmalloc/sizemap.h"

GOOGLE_MALLOC_SECTION_BEGIN
namespace tcmalloc {

namespace tcmalloc_internal {

// <fixed> is fixed per-size-class overhead due to end-of-span fragmentation
// and other factors. For instance, if we have a 96 byte size class, and use a
// single 8KiB page, then we will hold 85 objects per span, and have 32 bytes
// left over. There is also a fixed component of 48 bytes of TCMalloc metadata
// per span. Together, the fixed overhead would be wasted/allocated =
// (32 + 48) / (8192 - 32) ~= 0.98%.
// There is also a dynamic component to overhead based on mismatches between the
// number of bytes requested and the number of bytes provided by the size class.
// Together they sum to the total overhead; for instance if you asked for a
// 50-byte allocation that rounds up to a 64-byte size class, the dynamic
// overhead would be 28%, and if <fixed> were 22% it would mean (on average)
// 25 bytes of overhead for allocations of that size.

// clang-format off
#if defined(__cpp_aligned_new) && __STDCPP_DEFAULT_NEW_ALIGNMENT__ <= 8
#if TCMALLOC_PAGE_SHIFT == 13
static_assert(kMaxSize == 262144, "kMaxSize mismatch");
static const int kCount = 86;
static_assert(kCount <= kNumBaseClasses);
static constexpr SizeClassInfo kLegacySizeClassesList[kCount] = {
    // <bytes>, <pages>, <batch>, <capacity>    <fixed>
    {        0,       0,       0,          0},  // +Inf%
    {        8,       1,      32,       2281},  // 0.59%
    {       16,       1,      32,       2281},  // 0.59%
    {       24,       1,      32,       2281},  // 0.68%
    {       32,       1,      32,       2281},  // 0.59%
    {       40,       1,      32,       1398},  // 0.98%
    {       48,       1,      32,       2281},  // 0.98%
    {       56,       1,      32,        779},  // 0.78%
    {       64,       1,      32,       1554},  // 0.59%
    {       72,       1,      32,        565},  // 1.28%
    {       80,       1,      32,        841},  // 0.98%
    {       88,       1,      32,        476},  // 0.68%
    {       96,       1,      32,        539},  // 0.98%
    {      104,       1,      32,        276},  // 1.58%
    {      112,       1,      32,        392},  // 0.78%
    {      120,       1,      32,        282},  // 0.98%
    {      128,       1,      32,        448},  // 0.59%
    {      136,       1,      32,        226},  // 0.98%
    {      144,       1,      32,        334},  // 2.18%
    {      160,       1,      32,        361},  // 0.98%
    {      176,       1,      32,        235},  // 1.78%
    {      192,       1,      32,        277},  // 2.18%
    {      208,       1,      32,        216},  // 1.58%
    {      224,       1,      32,        200},  // 2.18%
    {      240,       1,      32,        194},  // 0.98%
    {      256,       1,      32,        322},  // 0.59%
    {      264,       1,      32,        154},  // 0.68%
    {      280,       1,      32,        204},  // 1.48%
    {      312,       1,      32,        224},  // 1.58%
    {      336,       1,      32,        228},  // 2.18%
    {      352,       1,      32,        165},  // 1.78%
    {      384,       1,      32,        186},  // 2.18%
    {      408,       1,      32,        174},  // 0.98%
    {      424,       1,      32,        156},  // 2.28%
    {      448,       1,      32,        182},  // 2.18%
    {      480,       1,      32,        167},  // 0.98%
    {      512,       1,      32,        246},  // 0.59%
    {      576,       1,      32,        188},  // 2.18%
    {      640,       1,      32,        175},  // 7.29%
    {      704,       1,      32,        165},  // 6.40%
    {      768,       1,      32,        164},  // 7.29%
    {      896,       1,      32,        170},  // 2.18%
    {     1024,       1,      32,        235},  // 0.59%
    {     1152,       2,      32,        167},  // 1.88%
    {     1280,       2,      32,        158},  // 6.98%
    {     1408,       2,      32,        155},  // 6.10%
    {     1536,       2,      32,        157},  // 6.98%
    {     1792,       2,      32,        156},  // 1.88%
    {     2048,       2,      32,        170},  // 0.29%
    {     2304,       2,      28,        154},  // 1.88%
    {     2688,       2,      24,        151},  // 1.88%
    {     2816,       3,      23,        146},  // 9.30%
    {     3200,       2,      20,        148},  // 2.70%
    {     3456,       3,      18,        145},  // 1.79%
    {     3584,       4,      18,        145},  // 1.74%
    {     4096,       1,      16,        225},  // 0.59%
    {     4736,       3,      13,        147},  // 3.99%
    {     5376,       2,      12,        145},  // 1.88%
    {     6144,       3,      10,        147},  // 0.20%
    {     7168,       7,       9,        146},  // 0.08%
    {     8192,       1,       8,        172},  // 0.59%
    {     9472,       5,       6,        146},  // 8.23%
    {    10240,       4,       6,        144},  // 6.82%
    {    12288,       3,       5,        146},  // 0.20%
    {    13568,       5,       4,        144},  // 0.75%
    {    14336,       7,       4,        143},  // 0.08%
    {    16384,       2,       4,        148},  // 0.29%
    {    20480,       5,       3,        145},  // 0.12%
    {    24576,       3,       2,        145},  // 0.20%
    {    28672,       7,       2,        144},  // 0.08%
    {    32768,       4,       2,        149},  // 0.15%
    {    40960,       5,       2,        144},  // 0.12%
    {    49152,       6,       2,        143},  // 0.10%
    {    57344,       7,       2,        143},  // 0.08%
    {    65536,       8,       2,        145},  // 0.07%
    {    73728,       9,       2,        144},  // 0.07%
    {    81920,      10,       2,        143},  // 0.06%
    {    98304,      12,       2,        144},  // 0.05%
    {   114688,      14,       2,        143},  // 0.04%
    {   131072,      16,       2,        143},  // 0.04%
    {   139264,      17,       2,        144},  // 0.03%
    {   155648,      19,       2,        143},  // 0.03%
    {   172032,      21,       2,        135},  // 0.03%
    {   204800,      25,       2,        143},  // 0.02%
    {   229376,      28,       2,        143},  // 0.02%
    {   262144,      32,       2,        143},  // 0.02%
};
constexpr absl::Span<const SizeClassInfo> kLegacySizeClasses(kLegacySizeClassesList);
#elif TCMALLOC_PAGE_SHIFT == 15
static_assert(kMaxSize == 262144, "kMaxSize mismatch");
static const int kCount = 78;
static_assert(kCount <= kNumBaseClasses);
static constexpr SizeClassInfo kLegacySizeClassesList[kCount] = {
    // <bytes>, <pages>, <batch>, <capacity>    <fixed>
    {        0,       0,       0,          0},  // +Inf%
    {        8,       1,      32,       2485},  // 0.15%
    {       16,       1,      32,       2485},  // 0.15%
    {       24,       1,      32,       2485},  // 0.17%
    {       32,       1,      32,       2485},  // 0.15%
    {       40,       1,      32,       1587},  // 0.17%
    {       48,       1,      32,       2485},  // 0.24%
    {       56,       1,      32,        781},  // 0.17%
    {       64,       1,      32,       1520},  // 0.15%
    {       72,       1,      32,        436},  // 0.17%
    {       80,       1,      32,        580},  // 0.29%
    {       88,       1,      32,        442},  // 0.24%
    {       96,       1,      32,        522},  // 0.24%
    {      104,       1,      32,        306},  // 0.17%
    {      112,       1,      32,        336},  // 0.34%
    {      120,       1,      32,        295},  // 0.17%
    {      128,       1,      32,        487},  // 0.15%
    {      136,       1,      32,        233},  // 0.54%
    {      144,       1,      32,        275},  // 0.39%
    {      160,       1,      32,        312},  // 0.54%
    {      176,       1,      32,        241},  // 0.24%
    {      192,       1,      32,        277},  // 0.54%
    {      208,       1,      32,        201},  // 0.49%
    {      240,       1,      32,        297},  // 0.54%
    {      256,       1,      32,        329},  // 0.15%
    {      280,       1,      32,        222},  // 0.17%
    {      304,       1,      32,        199},  // 0.89%
    {      320,       1,      32,        177},  // 0.54%
    {      352,       1,      32,        226},  // 0.24%
    {      400,       1,      32,        201},  // 1.28%
    {      448,       1,      32,        191},  // 0.34%
    {      512,       1,      32,        246},  // 0.15%
    {      576,       1,      32,        186},  // 1.74%
    {      640,       1,      32,        198},  // 0.54%
    {      704,       1,      32,        182},  // 1.33%
    {      768,       1,      32,        176},  // 1.74%
    {      896,       1,      32,        191},  // 1.74%
    {     1024,       1,      32,        217},  // 0.15%
    {     1152,       1,      32,        176},  // 1.74%
    {     1280,       1,      32,        170},  // 2.55%
    {     1408,       1,      32,        170},  // 1.33%
    {     1536,       1,      32,        167},  // 1.74%
    {     1792,       1,      32,        166},  // 1.74%
    {     1920,       1,      32,        159},  // 0.54%
    {     2048,       1,      32,        170},  // 0.15%
    {     2176,       1,      30,        167},  // 0.54%
    {     2304,       1,      28,        159},  // 1.74%
    {     2688,       1,      24,        165},  // 1.74%
    {     3200,       1,      20,        163},  // 2.55%
    {     3584,       1,      18,        159},  // 1.74%
    {     4096,       1,      16,        206},  // 0.15%
    {     4608,       1,      14,        160},  // 1.74%
    {     5376,       1,      12,        159},  // 1.74%
    {     6528,       1,      10,        163},  // 0.54%
    {     8192,       1,       8,        167},  // 0.15%
    {     9344,       2,       7,        159},  // 0.27%
    {    10880,       1,       6,        157},  // 0.54%
    {    13056,       2,       5,        158},  // 0.47%
    {    13952,       3,       4,        155},  // 0.70%
    {    16384,       1,       4,        164},  // 0.15%
    {    19072,       3,       3,        158},  // 3.14%
    {    21760,       2,       3,        157},  // 0.47%
    {    24576,       3,       2,        155},  // 0.05%
    {    28672,       7,       2,        157},  // 0.02%
    {    32768,       1,       2,        160},  // 0.15%
    {    38144,       5,       2,        155},  // 7.41%
    {    40960,       4,       2,        155},  // 6.71%
    {    49152,       3,       2,        155},  // 0.05%
    {    57344,       7,       2,        155},  // 0.02%
    {    65536,       2,       2,        158},  // 0.07%
    {    81920,       5,       2,        157},  // 0.03%
    {    98304,       3,       2,        155},  // 0.05%
    {   114688,       7,       2,        155},  // 0.02%
    {   131072,       4,       2,        163},  // 0.04%
    {   163840,       5,       2,        155},  // 0.03%
    {   196608,       6,       2,        155},  // 0.02%
    {   229376,       7,       2,        155},  // 0.02%
    {   262144,       8,       2,        157},  // 0.02%
};
constexpr absl::Span<const SizeClassInfo> kLegacySizeClasses(kLegacySizeClassesList);
#elif TCMALLOC_PAGE_SHIFT == 18
static_assert(kMaxSize == 262144, "kMaxSize mismatch");
static const int kCount = 89;
static_assert(kCount <= kNumBaseClasses);
static constexpr SizeClassInfo kLegacySizeClassesList[kCount] = {
    // <bytes>, <pages>, <batch>, <capacity>    <fixed>
    {        0,       0,       0,          0},  // +Inf%
    {        8,       1,      32,       2331},  // 0.02%
    {       16,       1,      32,       2331},  // 0.02%
    {       24,       1,      32,       2331},  // 0.02%
    {       32,       1,      32,       2331},  // 0.02%
    {       40,       1,      32,       1334},  // 0.03%
    {       48,       1,      32,       2331},  // 0.02%
    {       56,       1,      32,        648},  // 0.02%
    {       64,       1,      32,       1459},  // 0.02%
    {       72,       1,      32,        691},  // 0.04%
    {       80,       1,      32,        624},  // 0.04%
    {       88,       1,      32,        302},  // 0.05%
    {       96,       1,      32,        382},  // 0.04%
    {      104,       1,      32,        264},  // 0.04%
    {      112,       1,      32,        340},  // 0.04%
    {      128,       1,      32,        473},  // 0.02%
    {      144,       1,      32,        411},  // 0.04%
    {      160,       1,      32,        365},  // 0.04%
    {      176,       1,      32,        215},  // 0.05%
    {      192,       1,      32,        236},  // 0.04%
    {      208,       1,      32,        200},  // 0.04%
    {      232,       1,      32,        299},  // 0.10%
    {      256,       1,      32,        333},  // 0.02%
    {      280,       1,      32,        221},  // 0.04%
    {      312,       1,      32,        273},  // 0.04%
    {      336,       1,      32,        246},  // 0.04%
    {      376,       1,      32,        176},  // 0.05%
    {      416,       1,      32,        199},  // 0.04%
    {      472,       1,      32,        193},  // 0.09%
    {      512,       1,      32,        213},  // 0.02%
    {      576,       1,      32,        190},  // 0.04%
    {      704,       1,      32,        207},  // 0.12%
    {      768,       1,      32,        165},  // 0.12%
    {      896,       1,      32,        174},  // 0.21%
    {     1024,       1,      32,        213},  // 0.02%
    {     1152,       1,      32,        168},  // 0.26%
    {     1280,       1,      32,        162},  // 0.41%
    {     1408,       1,      32,        155},  // 0.12%
    {     1664,       1,      32,        188},  // 0.36%
    {     1920,       1,      32,        174},  // 0.41%
    {     2048,       1,      32,        165},  // 0.02%
    {     2176,       1,      30,        204},  // 0.41%
    {     2304,       1,      28,        156},  // 0.71%
    {     2560,       1,      25,        154},  // 0.41%
    {     2816,       1,      23,        149},  // 0.12%
    {     3072,       1,      21,        149},  // 0.41%
    {     3328,       1,      19,        154},  // 1.00%
    {     3584,       1,      18,        148},  // 0.21%
    {     3840,       1,      17,        148},  // 0.41%
    {     4096,       1,      16,        192},  // 0.02%
    {     4224,       1,      15,        149},  // 0.12%
    {     4736,       1,      13,        151},  // 0.66%
    {     5248,       1,      12,        154},  // 1.96%
    {     5760,       1,      11,        148},  // 1.15%
    {     6528,       1,      10,        150},  // 0.41%
    {     7168,       1,       9,        147},  // 1.61%
    {     8192,       1,       8,        160},  // 0.02%
    {     9344,       1,       7,        149},  // 0.21%
    {    10880,       1,       6,        148},  // 0.41%
    {    11904,       1,       5,        148},  // 0.12%
    {    13056,       1,       5,        148},  // 0.41%
    {    13696,       1,       4,        146},  // 0.76%
    {    14464,       1,       4,        147},  // 0.71%
    {    15360,       1,       4,        146},  // 0.41%
    {    16384,       1,       4,        151},  // 0.02%
    {    17408,       1,       3,        147},  // 0.41%
    {    18688,       1,       3,        148},  // 0.21%
    {    20096,       1,       3,        146},  // 0.36%
    {    21760,       1,       3,        146},  // 0.41%
    {    23808,       1,       2,        147},  // 0.12%
    {    26112,       1,       2,        147},  // 0.41%
    {    29056,       1,       2,        146},  // 0.26%
    {    32768,       1,       2,        162},  // 0.02%
    {    37376,       1,       2,        147},  // 0.21%
    {    43648,       1,       2,        146},  // 0.12%
    {    45568,       2,       2,        146},  // 4.61%
    {    52352,       1,       2,        146},  // 0.17%
    {    56064,       2,       2,        146},  // 3.92%
    {    65536,       1,       2,        147},  // 0.02%
    {    74880,       2,       2,        146},  // 0.03%
    {    87296,       1,       2,        146},  // 0.12%
    {   104832,       2,       2,        146},  // 0.03%
    {   112256,       3,       2,        146},  // 0.09%
    {   131072,       1,       2,        146},  // 0.02%
    {   149760,       3,       2,        146},  // 5.03%
    {   174720,       2,       2,        146},  // 0.03%
    {   196608,       3,       2,        137},  // 0.01%
    {   209664,       4,       2,        146},  // 0.03%
    {   262144,       1,       2,        147},  // 0.02%
};
constexpr absl::Span<const SizeClassInfo> kLegacySizeClasses(kLegacySizeClassesList);
#elif TCMALLOC_PAGE_SHIFT == 12
static_assert(kMaxSize == 8192, "kMaxSize mismatch");
static const int kCount = 46;
static_assert(kCount <= kNumBaseClasses);
static constexpr SizeClassInfo kLegacySizeClassesList[kCount] = {
    // <bytes>, <pages>, <batch>, <capacity>    <fixed>
    {        0,       0,       0,          0},  // +Inf%
    {        8,       1,      32,       2072},  // 1.17%
    {       16,       1,      32,       2810},  // 1.17%
    {       24,       1,      32,       2810},  // 1.57%
    {       32,       1,      32,       2810},  // 1.17%
    {       40,       1,      32,       2557},  // 1.57%
    {       48,       1,      32,       2810},  // 1.57%
    {       56,       1,      32,       1180},  // 1.37%
    {       64,       1,      32,       2810},  // 1.17%
    {       72,       1,      32,        475},  // 2.78%
    {       80,       1,      32,       1139},  // 1.57%
    {       96,       1,      32,        958},  // 2.78%
    {      104,       1,      32,        375},  // 2.17%
    {      112,       1,      32,        580},  // 2.78%
    {      128,       1,      32,        741},  // 1.17%
    {      144,       1,      32,        617},  // 2.78%
    {      160,       1,      32,        568},  // 3.60%
    {      176,       1,      32,        332},  // 2.37%
    {      192,       1,      32,        350},  // 2.78%
    {      208,       1,      32,        266},  // 4.86%
    {      224,       1,      32,        327},  // 2.78%
    {      240,       1,      32,        236},  // 1.57%
    {      256,       1,      32,        346},  // 1.17%
    {      272,       1,      32,        213},  // 1.57%
    {      288,       1,      32,        229},  // 2.78%
    {      336,       1,      32,        339},  // 2.78%
    {      408,       1,      32,        210},  // 1.57%
    {      448,       1,      32,        200},  // 2.78%
    {      512,       1,      32,        401},  // 1.17%
    {      576,       2,      32,        244},  // 2.18%
    {      640,       2,      32,        185},  // 7.29%
    {      768,       2,      32,        243},  // 7.29%
    {      896,       2,      32,        224},  // 2.18%
    {     1024,       2,      32,        488},  // 0.59%
    {     1152,       3,      32,        193},  // 7.08%
    {     1280,       3,      32,        176},  // 7.08%
    {     1536,       3,      32,        191},  // 0.39%
    {     2048,       4,      32,        318},  // 0.29%
    {     2304,       4,      28,        187},  // 1.88%
    {     2688,       4,      24,        182},  // 1.88%
    {     3200,       4,      20,        177},  // 2.70%
    {     4096,       4,      16,        355},  // 0.29%
    {     4736,       5,      13,        195},  // 8.36%
    {     6144,       3,      10,        177},  // 0.39%
    {     7168,       7,       9,        178},  // 0.17%
    {     8192,       4,       8,        204},  // 0.29%
};
constexpr absl::Span<const SizeClassInfo> kLegacySizeClasses(kLegacySizeClassesList);
#else
#error "Unsupported TCMALLOC_PAGE_SHIFT value!"
#endif
#else
#if TCMALLOC_PAGE_SHIFT == 13
static_assert(kMaxSize == 262144, "kMaxSize mismatch");
static const int kCount = 86;
static_assert(kCount <= kNumBaseClasses);
static constexpr SizeClassInfo kLegacySizeClassesList[kCount] = {
    // <bytes>, <pages>, <batch>, <capacity>    <fixed>
    {        0,       0,       0,          0},  // +Inf%
    {        8,       1,      32,       2530},  // 0.59%
    {       16,       1,      32,       2530},  // 0.59%
    {       32,       1,      32,       2534},  // 0.59%
    {       48,       1,      32,       2530},  // 0.98%
    {       64,       1,      32,       2530},  // 0.59%
    {       80,       1,      32,       1459},  // 0.98%
    {       96,       1,      32,       1007},  // 0.98%
    {      112,       1,      32,        605},  // 0.78%
    {      128,       1,      32,        678},  // 0.59%
    {      144,       1,      32,        481},  // 2.18%
    {      160,       1,      32,        415},  // 0.98%
    {      176,       1,      32,        267},  // 1.78%
    {      192,       1,      32,        316},  // 2.18%
    {      208,       1,      32,        245},  // 1.58%
    {      224,       1,      32,        226},  // 2.18%
    {      240,       1,      32,        219},  // 0.98%
    {      256,       1,      32,        368},  // 0.59%
    {      272,       1,      32,        206},  // 0.98%
    {      288,       1,      32,        226},  // 2.18%
    {      304,       1,      32,        191},  // 4.25%
    {      320,       1,      32,        217},  // 3.00%
    {      336,       1,      32,        230},  // 2.18%
    {      352,       1,      32,        184},  // 1.78%
    {      368,       1,      32,        173},  // 1.78%
    {      384,       1,      32,        194},  // 2.18%
    {      400,       1,      32,        183},  // 3.00%
    {      416,       1,      32,        180},  // 4.25%
    {      448,       1,      32,        210},  // 2.18%
    {      480,       1,      32,        188},  // 0.98%
    {      512,       1,      32,        280},  // 0.59%
    {      576,       1,      32,        212},  // 2.18%
    {      640,       1,      32,        196},  // 7.29%
    {      704,       1,      32,        184},  // 6.40%
    {      768,       1,      32,        183},  // 7.29%
    {      896,       1,      32,        191},  // 2.18%
    {     1024,       1,      32,        266},  // 0.59%
    {     1152,       2,      32,        187},  // 1.88%
    {     1280,       2,      32,        177},  // 6.98%
    {     1408,       2,      32,        173},  // 6.10%
    {     1536,       2,      32,        175},  // 6.98%
    {     1792,       2,      32,        174},  // 1.88%
    {     2048,       2,      32,        190},  // 0.29%
    {     2304,       2,      28,        172},  // 1.88%
    {     2688,       2,      24,        168},  // 1.88%
    {     2816,       3,      23,        162},  // 9.30%
    {     3200,       2,      20,        164},  // 2.70%
    {     3456,       3,      18,        161},  // 1.79%
    {     3584,       4,      18,        161},  // 1.74%
    {     4096,       1,      16,        254},  // 0.59%
    {     4736,       3,      13,        164},  // 3.99%
    {     5376,       2,      12,        161},  // 1.88%
    {     6144,       3,      10,        164},  // 0.20%
    {     6528,       4,      10,        159},  // 0.54%
    {     7168,       7,       9,        161},  // 0.08%
    {     8192,       1,       8,        193},  // 0.59%
    {     9472,       5,       6,        162},  // 8.23%
    {    10240,       4,       6,        159},  // 6.82%
    {    12288,       3,       5,        162},  // 0.20%
    {    13568,       5,       4,        159},  // 0.75%
    {    14336,       7,       4,        158},  // 0.08%
    {    16384,       2,       4,        164},  // 0.29%
    {    20480,       5,       3,        161},  // 0.12%
    {    24576,       3,       2,        161},  // 0.20%
    {    28672,       7,       2,        159},  // 0.08%
    {    32768,       4,       2,        166},  // 0.15%
    {    40960,       5,       2,        159},  // 0.12%
    {    49152,       6,       2,        158},  // 0.10%
    {    57344,       7,       2,        159},  // 0.08%
    {    65536,       8,       2,        161},  // 0.07%
    {    73728,       9,       2,        159},  // 0.07%
    {    81920,      10,       2,        159},  // 0.06%
    {    90112,      11,       2,        158},  // 0.05%
    {    98304,      12,       2,        159},  // 0.05%
    {   106496,      13,       2,        158},  // 0.05%
    {   114688,      14,       2,        158},  // 0.04%
    {   131072,      16,       2,        158},  // 0.04%
    {   139264,      17,       2,        159},  // 0.03%
    {   147456,      18,       2,        158},  // 0.03%
    {   155648,      19,       2,        158},  // 0.03%
    {   172032,      21,       2,        158},  // 0.03%
    {   188416,      23,       2,        158},  // 0.03%
    {   204800,      25,       2,        158},  // 0.02%
    {   221184,      27,       2,        158},  // 0.02%
    {   237568,      29,       2,        158},  // 0.02%
    {   262144,      32,       2,        158},  // 0.02%
};
constexpr absl::Span<const SizeClassInfo> kLegacySizeClasses(kLegacySizeClassesList);
#elif TCMALLOC_PAGE_SHIFT == 15
static_assert(kMaxSize == 262144, "kMaxSize mismatch");
static const int kCount = 78;
static_assert(kCount <= kNumBaseClasses);
static constexpr SizeClassInfo kLegacySizeClassesList[kCount] = {
    // <bytes>, <pages>, <batch>, <capacity>    <fixed>
    {        0,       0,       0,          0},  // +Inf%
    {        8,       1,      32,       2724},  // 0.15%
    {       16,       1,      32,       2724},  // 0.15%
    {       32,       1,      32,       2724},  // 0.15%
    {       48,       1,      32,       2724},  // 0.24%
    {       64,       1,      32,       2724},  // 0.15%
    {       80,       1,      32,       1075},  // 0.29%
    {       96,       1,      32,       1011},  // 0.24%
    {      112,       1,      32,        599},  // 0.34%
    {      128,       1,      32,        781},  // 0.15%
    {      144,       1,      32,        427},  // 0.39%
    {      160,       1,      32,        374},  // 0.54%
    {      176,       1,      32,        282},  // 0.24%
    {      192,       1,      32,        329},  // 0.54%
    {      208,       1,      32,        230},  // 0.49%
    {      224,       1,      32,        263},  // 0.34%
    {      240,       1,      32,        262},  // 0.54%
    {      256,       1,      32,        395},  // 0.15%
    {      288,       1,      32,        274},  // 0.84%
    {      304,       1,      32,        210},  // 0.89%
    {      320,       1,      32,        200},  // 0.54%
    {      352,       1,      32,        262},  // 0.24%
    {      384,       1,      32,        214},  // 0.54%
    {      400,       1,      32,        185},  // 1.28%
    {      448,       1,      32,        215},  // 0.34%
    {      480,       1,      32,        218},  // 0.54%
    {      512,       1,      32,        241},  // 0.15%
    {      576,       1,      32,        210},  // 1.74%
    {      640,       1,      32,        225},  // 0.54%
    {      704,       1,      32,        205},  // 1.33%
    {      768,       1,      32,        197},  // 1.74%
    {      832,       1,      32,        198},  // 1.13%
    {      896,       1,      32,        188},  // 1.74%
    {     1024,       1,      32,        251},  // 0.15%
    {     1152,       1,      32,        197},  // 1.74%
    {     1280,       1,      32,        190},  // 2.55%
    {     1408,       1,      32,        190},  // 1.33%
    {     1536,       1,      32,        186},  // 1.74%
    {     1792,       1,      32,        185},  // 1.74%
    {     1920,       1,      32,        174},  // 0.54%
    {     2048,       1,      32,        189},  // 0.15%
    {     2176,       1,      30,        186},  // 0.54%
    {     2304,       1,      28,        176},  // 1.74%
    {     2432,       1,      26,        174},  // 3.80%
    {     2688,       1,      24,        178},  // 1.74%
    {     2944,       1,      22,        174},  // 1.33%
    {     3200,       1,      20,        176},  // 2.55%
    {     3584,       1,      18,        174},  // 1.74%
    {     4096,       1,      16,        237},  // 0.15%
    {     4608,       1,      14,        177},  // 1.74%
    {     5376,       1,      12,        174},  // 1.74%
    {     6528,       1,      10,        180},  // 0.54%
    {     7168,       2,       9,        172},  // 1.66%
    {     8192,       1,       8,        186},  // 0.15%
    {     9344,       2,       7,        174},  // 0.27%
    {    10880,       1,       6,        172},  // 0.54%
    {    13056,       2,       5,        173},  // 0.47%
    {    13952,       3,       4,        170},  // 0.70%
    {    16384,       1,       4,        181},  // 0.15%
    {    19072,       3,       3,        174},  // 3.14%
    {    21760,       2,       3,        172},  // 0.47%
    {    24576,       3,       2,        172},  // 0.05%
    {    26112,       4,       2,        170},  // 0.43%
    {    28672,       7,       2,        172},  // 0.02%
    {    32768,       1,       2,        177},  // 0.15%
    {    38144,       5,       2,        172},  // 7.41%
    {    40960,       4,       2,        170},  // 6.71%
    {    49152,       3,       2,        170},  // 0.05%
    {    57344,       7,       2,        172},  // 0.02%
    {    65536,       2,       2,        173},  // 0.07%
    {    81920,       5,       2,        172},  // 0.03%
    {    98304,       3,       2,        170},  // 0.05%
    {   114688,       7,       2,        170},  // 0.02%
    {   131072,       4,       2,        180},  // 0.04%
    {   163840,       5,       2,        170},  // 0.03%
    {   196608,       6,       2,        170},  // 0.02%
    {   229376,       7,       2,        166},  // 0.02%
    {   262144,       8,       2,        172},  // 0.02%
};
constexpr absl::Span<const SizeClassInfo> kLegacySizeClasses(kLegacySizeClassesList);
#elif TCMALLOC_PAGE_SHIFT == 18
static_assert(kMaxSize == 262144, "kMaxSize mismatch");
static const int kCount = 89;
static_assert(kCount <= kNumBaseClasses);
static constexpr SizeClassInfo kLegacySizeClassesList[kCount] = {
    // <bytes>, <pages>, <batch>, <capacity>    <fixed>
    {        0,       0,       0,          0},  // +Inf%
    {        8,       1,      32,       2502},  // 0.02%
    {       16,       1,      32,       2502},  // 0.02%
    {       32,       1,      32,       2506},  // 0.02%
    {       48,       1,      32,       2502},  // 0.02%
    {       64,       1,      32,       2502},  // 0.02%
    {       80,       1,      32,       1473},  // 0.04%
    {       96,       1,      32,        666},  // 0.04%
    {      112,       1,      32,        563},  // 0.04%
    {      128,       1,      32,        585},  // 0.02%
    {      144,       1,      32,        503},  // 0.04%
    {      160,       1,      32,        443},  // 0.04%
    {      176,       1,      32,        248},  // 0.05%
    {      192,       1,      32,        275},  // 0.04%
    {      208,       1,      32,        227},  // 0.04%
    {      224,       1,      32,        345},  // 0.04%
    {      256,       1,      32,        415},  // 0.02%
    {      288,       1,      32,        334},  // 0.04%
    {      320,       1,      32,        271},  // 0.04%
    {      336,       1,      32,        259},  // 0.04%
    {      368,       1,      32,        187},  // 0.07%
    {      400,       1,      32,        221},  // 0.07%
    {      448,       1,      32,        222},  // 0.04%
    {      480,       1,      32,        181},  // 0.04%
    {      512,       1,      32,        232},  // 0.02%
    {      576,       1,      32,        215},  // 0.04%
    {      640,       1,      32,        208},  // 0.17%
    {      704,       1,      32,        184},  // 0.12%
    {      768,       1,      32,        181},  // 0.12%
    {      896,       1,      32,        193},  // 0.21%
    {     1024,       1,      32,        244},  // 0.02%
    {     1152,       1,      32,        186},  // 0.26%
    {     1280,       1,      32,        177},  // 0.41%
    {     1408,       1,      32,        167},  // 0.12%
    {     1536,       1,      32,        169},  // 0.41%
    {     1664,       1,      32,        199},  // 0.36%
    {     1920,       1,      32,        193},  // 0.41%
    {     2048,       1,      32,        182},  // 0.02%
    {     2176,       1,      30,        233},  // 0.41%
    {     2304,       1,      28,        170},  // 0.71%
    {     2560,       1,      25,        167},  // 0.41%
    {     2816,       1,      23,        160},  // 0.12%
    {     3072,       1,      21,        160},  // 0.41%
    {     3328,       1,      19,        167},  // 1.00%
    {     3584,       1,      18,        159},  // 0.21%
    {     3840,       1,      17,        159},  // 0.41%
    {     4096,       1,      16,        217},  // 0.02%
    {     4224,       1,      15,        161},  // 0.12%
    {     4736,       1,      13,        164},  // 0.66%
    {     5120,       1,      12,        162},  // 0.41%
    {     5632,       1,      11,        164},  // 1.20%
    {     6144,       1,      10,        159},  // 1.61%
    {     6528,       1,      10,        160},  // 0.41%
    {     7168,       1,       9,        158},  // 1.61%
    {     8192,       1,       8,        176},  // 0.02%
    {     8704,       1,       7,        158},  // 0.41%
    {     9344,       1,       7,        159},  // 0.21%
    {    10368,       1,       6,        158},  // 1.15%
    {    11392,       1,       5,        160},  // 0.07%
    {    12416,       1,       5,        159},  // 0.56%
    {    13056,       1,       5,        158},  // 0.41%
    {    13696,       1,       4,        156},  // 0.76%
    {    14464,       1,       4,        158},  // 0.71%
    {    15360,       1,       4,        158},  // 0.41%
    {    16384,       1,       4,        164},  // 0.02%
    {    17408,       1,       3,        158},  // 0.41%
    {    18688,       1,       3,        159},  // 0.21%
    {    20096,       1,       3,        156},  // 0.36%
    {    21760,       1,       3,        158},  // 0.41%
    {    23808,       1,       2,        159},  // 0.12%
    {    26112,       1,       2,        158},  // 0.41%
    {    29056,       1,       2,        156},  // 0.26%
    {    32768,       1,       2,        176},  // 0.02%
    {    37376,       1,       2,        158},  // 0.21%
    {    43648,       1,       2,        156},  // 0.12%
    {    45568,       2,       2,        156},  // 4.61%
    {    52352,       1,       2,        156},  // 0.17%
    {    56064,       2,       2,        156},  // 3.92%
    {    65536,       1,       2,        158},  // 0.02%
    {    74880,       2,       2,        156},  // 0.03%
    {    87296,       1,       2,        156},  // 0.12%
    {   104832,       2,       2,        156},  // 0.03%
    {   112256,       3,       2,        156},  // 0.09%
    {   131072,       1,       2,        156},  // 0.02%
    {   149760,       3,       2,        156},  // 5.03%
    {   174720,       2,       2,        156},  // 0.03%
    {   196608,       3,       2,        156},  // 0.01%
    {   209664,       4,       2,        156},  // 0.03%
    {   262144,       1,       2,        158},  // 0.02%
};
constexpr absl::Span<const SizeClassInfo> kLegacySizeClasses(kLegacySizeClassesList);
#elif TCMALLOC_PAGE_SHIFT == 12
static_assert(kMaxSize == 8192, "kMaxSize mismatch");
static const int kCount = 46;
static_assert(kCount <= kNumBaseClasses);
static constexpr SizeClassInfo kLegacySizeClassesList[kCount] = {
    // <bytes>, <pages>, <batch>, <capacity>    <fixed>
    {        0,       0,       0,          0},  // +Inf%
    {        8,       1,      32,       3071},  // 1.17%
    {       16,       1,      32,       3071},  // 1.17%
    {       32,       1,      32,       3078},  // 1.17%
    {       48,       1,      32,       3071},  // 1.57%
    {       64,       1,      32,       3071},  // 1.17%
    {       80,       1,      32,       2098},  // 1.57%
    {       96,       1,      32,       1380},  // 2.78%
    {      112,       1,      32,       1108},  // 2.78%
    {      128,       1,      32,       1051},  // 1.17%
    {      144,       1,      32,        865},  // 2.78%
    {      160,       1,      32,        790},  // 3.60%
    {      176,       1,      32,        432},  // 2.37%
    {      192,       1,      32,        459},  // 2.78%
    {      208,       1,      32,        331},  // 4.86%
    {      224,       1,      32,        423},  // 2.78%
    {      240,       1,      32,        286},  // 1.57%
    {      256,       1,      32,        451},  // 1.17%
    {      272,       1,      32,        252},  // 1.57%
    {      288,       1,      32,        276},  // 2.78%
    {      304,       1,      32,        247},  // 4.86%
    {      336,       1,      32,        384},  // 2.78%
    {      368,       1,      32,        223},  // 2.37%
    {      400,       1,      32,        210},  // 3.60%
    {      448,       1,      32,        234},  // 2.78%
    {      512,       1,      32,        535},  // 1.17%
    {      576,       2,      32,        298},  // 2.18%
    {      640,       2,      32,        208},  // 7.29%
    {      768,       2,      32,        295},  // 7.29%
    {      896,       2,      32,        267},  // 2.18%
    {     1024,       2,      32,        669},  // 0.59%
    {     1152,       3,      32,        222},  // 7.08%
    {     1280,       3,      32,        193},  // 7.08%
    {     1536,       3,      32,        217},  // 0.39%
    {     1792,       4,      32,        199},  // 1.88%
    {     2048,       4,      32,        402},  // 0.29%
    {     2304,       4,      28,        210},  // 1.88%
    {     2688,       4,      24,        204},  // 1.88%
    {     3200,       4,      20,        196},  // 2.70%
    {     3584,       7,      18,        193},  // 0.17%
    {     4096,       4,      16,        463},  // 0.29%
    {     4736,       5,      13,        223},  // 8.36%
    {     5376,       4,      12,        192},  // 1.88%
    {     6144,       3,      10,        195},  // 0.39%
    {     7168,       7,       9,        198},  // 0.17%
    {     8192,       4,       8,        237},  // 0.29%
};
constexpr absl::Span<const SizeClassInfo> kLegacySizeClasses(kLegacySizeClassesList);
#else
#error "Unsupported TCMALLOC_PAGE_SHIFT value!"
#endif
#endif
// clang-format on

}  // namespace tcmalloc_internal
}  // namespace tcmalloc
GOOGLE_MALLOC_SECTION_END
